@@ -3,7 +3,7 @@
  */
 
 #include <stdio.h>
-
+#include <stddef.h>
 #include <ctype.h>
 #include <signal.h>
 #include <setjmp.h>
@@ -12,25 +12,19 @@
 #include "ntp_select.h"
 #include "ntp_io.h"
 #include "ntp_stdlib.h"
-/* Don't include ISC's version of IPv6 variables and structures */
-#define ISC_IPV6_H 1
+#include "ntp_assert.h"
+#include "ntp_lineedit.h"
 #include "isc/net.h"
 #include "isc/result.h"
+#include <ssl_applink.c>
 
 #include "ntpdc-opts.h"
 
 #ifdef SYS_WINNT
 # include <Mswsock.h>
 # include <io.h>
-#else
-# define closesocket close
 #endif /* SYS_WINNT */
 
-#if defined(HAVE_LIBREADLINE) || defined (HAVE_LIBEDIT)
-# include <readline/readline.h>
-# include <readline/history.h>
-#endif /* HAVE_LIBREADLINE || HAVE_LIBEDIT */
-
 #ifdef SYS_VXWORKS
 				/* vxWorks needs mode flag -casey*/
 # define open(name, flags)   open(name, flags, 0777)
@@ -56,12 +50,8 @@ static	const char *	prompt = "ntpdc> ";	/* prompt to ask him about */
 static	u_long	info_auth_keyid;
 static int keyid_entered = 0;
 
-/*
- * Type of key md5
- */
-#define	KEY_TYPE_MD5	4
-
-static	int info_auth_keytype = KEY_TYPE_MD5;	/* MD5 */
+static	int	info_auth_keytype = NID_md5;	/* MD5 */
+static	size_t	info_auth_hashlen = 16;		/* MD5 */
 u_long	current_time;		/* needed by authkeys; not used */
 
 /*
@@ -69,42 +59,50 @@ u_long	current_time;		/* needed by authkeys; not used */
  */
 s_char	sys_precision;		/* local clock precision (log2 s) */
 
-int		ntpdcmain	P((int,	char **));
+/*
+ * Use getpassphrase() if configure.ac detected it, as Suns that
+ * have it truncate the password in getpass() to 8 characters.
+ */
+#ifdef HAVE_GETPASSPHRASE
+# define	getpass(str)	getpassphrase(str)
+#endif
+
+int		ntpdcmain	(int,	char **);
 /*
  * Built in command handler declarations
  */
-static	int	openhost	P((const char *));
-static	int	sendpkt		P((char *, int));
-static	void	growpktdata	P((void));
-static	int	getresponse	P((int, int, int *, int *, char **, int));
-static	int	sendrequest	P((int, int, int, int, int, char *));
-static	void	getcmds		P((void));
-static	RETSIGTYPE abortcmd	P((int));
-static	void	docmd		P((const char *));
-static	void	tokenize	P((const char *, char **, int *));
-static	int	findcmd		P((char *, struct xcmd *, struct xcmd *, struct xcmd **));
-static	int	getarg		P((char *, int, arg_v *));
-static	int	getnetnum	P((const char *, struct sockaddr_storage *, char *, int));
-static	void	help		P((struct parse *, FILE *));
+static	int	openhost	(const char *);
+static	int	sendpkt		(void *, size_t);
+static	void	growpktdata	(void);
+static	int	getresponse	(int, int, int *, int *, char **, int);
+static	int	sendrequest	(int, int, int, u_int, size_t, char *);
+static	void	getcmds		(void);
+static	RETSIGTYPE abortcmd	(int);
+static	void	docmd		(const char *);
+static	void	tokenize	(const char *, char **, int *);
+static	int	findcmd		(char *, struct xcmd *, struct xcmd *, struct xcmd **);
+static	int	getarg		(char *, int, arg_v *);
+static	int	getnetnum	(const char *, sockaddr_u *, char *, int);
+static	void	help		(struct parse *, FILE *);
 #ifdef QSORT_USES_VOID_P
-static	int	helpsort	P((const void *, const void *));
+static	int	helpsort	(const void *, const void *);
 #else
-static	int	helpsort	P((char **, char **));
+static	int	helpsort	(char **, char **);
 #endif
-static	void	printusage	P((struct xcmd *, FILE *));
-static	void	timeout		P((struct parse *, FILE *));
-static	void	my_delay	P((struct parse *, FILE *));
-static	void	host		P((struct parse *, FILE *));
-static	void	keyid		P((struct parse *, FILE *));
-static	void	keytype		P((struct parse *, FILE *));
-static	void	passwd		P((struct parse *, FILE *));
-static	void	hostnames	P((struct parse *, FILE *));
-static	void	setdebug	P((struct parse *, FILE *));
-static	void	quit		P((struct parse *, FILE *));
-static	void	version		P((struct parse *, FILE *));
-static	void	warning		P((const char *, const char *, const char *));
-static	void	error		P((const char *, const char *, const char *));
-static	u_long	getkeyid	P((const char *));
+static	void	printusage	(struct xcmd *, FILE *);
+static	void	timeout		(struct parse *, FILE *);
+static	void	my_delay	(struct parse *, FILE *);
+static	void	host		(struct parse *, FILE *);
+static	void	keyid		(struct parse *, FILE *);
+static	void	keytype		(struct parse *, FILE *);
+static	void	passwd		(struct parse *, FILE *);
+static	void	hostnames	(struct parse *, FILE *);
+static	void	setdebug	(struct parse *, FILE *);
+static	void	quit		(struct parse *, FILE *);
+static	void	version		(struct parse *, FILE *);
+static	void	warning		(const char *, const char *, const char *);
+static	void	error		(const char *, const char *, const char *);
+static	u_long	getkeyid	(const char *);
 
 
 
@@ -159,10 +157,10 @@ static	struct xcmd builtins[] = {
 /*
  * Default values we use.
  */
+#define	DEFHOST		"localhost"	/* default host name */
 #define	DEFTIMEOUT	(5)		/* 5 second time out */
 #define	DEFSTIMEOUT	(2)		/* 2 second time out after first */
 #define	DEFDELAY	0x51EB852	/* 20 milliseconds, l_fp fraction */
-#define	DEFHOST		"localhost"	/* default host name */
 #define	LENHOSTNAME	256		/* host name is 256 characters long */
 #define	MAXCMDS		100		/* maximum commands on cmd line */
 #define	MAXHOSTS	200		/* maximum hosts on cmd line */
@@ -173,8 +171,8 @@ static	struct xcmd builtins[] = {
 /*
  * Some variables used and manipulated locally
  */
-static	struct timeval tvout = { DEFTIMEOUT, 0 };	/* time out for reads */
-static	struct timeval tvsout = { DEFSTIMEOUT, 0 };	/* secondary time out */
+static	struct sock_timeval tvout = { DEFTIMEOUT, 0 };	/* time out for reads */
+static	struct sock_timeval tvsout = { DEFSTIMEOUT, 0 };/* secondary time out */
 static	l_fp delay_time;				/* delay time */
 static	char currenthost[LENHOSTNAME];			/* current host name */
 int showhostnames = 1;					/* show host names by default */
@@ -185,18 +183,6 @@ static	SOCKET sockfd;					/* fd socket is opened on */
 static	int havehost = 0;				/* set to 1 when host open */
 int s_port = 0;
 
-#if defined (SYS_WINNT) || defined (SYS_VXWORKS)
-char password[9];
-#endif /* SYS_WINNT || SYS_VXWORKS */
-
-#ifdef SYS_WINNT
-DWORD NumberOfBytesWritten;
-
-HANDLE	TimerThreadHandle = NULL;	/* 1998/06/03 - Used in ntplib/machines.c */
-void timer(void)	{  ; };	/* 1998/06/03 - Used in ntplib/machines.c */
-
-#endif /* SYS_WINNT */
-
 /*
  * Holds data returned from queries.  We allocate INITDATASIZE
  * octets to begin with, increasing this as we need to.
@@ -306,18 +292,12 @@ ntpdcmain(
 	taskPrioritySet(taskIdSelf(), 100 );
 #endif
 
-#ifdef SYS_WINNT
-	if (!Win32InitSockets())
-	{
-		fprintf(stderr, "No useable winsock.dll:");
-		exit(1);
-	}
-#endif /* SYS_WINNT */
+	init_lib();	/* sets up ipv4_works, ipv6_works */
+	ssl_applink();
 
-	/* Check to see if we have IPv6. Otherwise force the -4 flag */
-	if (isc_net_probeipv6() != ISC_R_SUCCESS) {
+	/* Check to see if we have IPv6. Otherwise default to IPv4 */
+	if (!ipv6_works)
 		ai_fam_default = AF_INET;
-	}
 
 	progname = argv[0];
 
@@ -446,12 +426,8 @@ ntpdcmain(
 	/*
 	 * Initialize the packet data buffer
 	 */
-	pktdata = (char *)malloc(INITDATASIZE);
-	if (pktdata == NULL) {
-		(void) fprintf(stderr, "%s: malloc() failed!\n", progname);
-		exit(1);
-	}
 	pktdatasize = INITDATASIZE;
+	pktdata = emalloc(INITDATASIZE);
 
 	if (numcmds == 0) {
 		(void) openhost(chosts[0]);
@@ -499,10 +475,14 @@ openhost(
 	
 	if (*cp == '[') {
 		cp++;	
-		for(i = 0; *cp != ']'; cp++, i++)
-			name[i] = *cp;	
-		name[i] = '\0';
-		hname = name;
+		for (i = 0; *cp && *cp != ']'; cp++, i++)
+			name[i] = *cp;
+		if (*cp == ']') {
+			name[i] = '\0';
+			hname = name;
+		} else {
+			return 0;
+		}
 	}	
 
 	/*
@@ -543,8 +523,14 @@ openhost(
 		return 0;
 	}
 
+	/* 
+	 * getaddrinfo() has returned without error so ai should not 
+	 * be NULL.
+	 */
+	NTP_INSIST(ai != NULL);
+
 	if (ai->ai_canonname == NULL) {
-		strncpy(temphost, stoa((struct sockaddr_storage *)ai->ai_addr),
+		strncpy(temphost, stoa((sockaddr_u *)ai->ai_addr),
 		    LENHOSTNAME);
 		temphost[LENHOSTNAME-1] = '\0';
 	} else {
@@ -619,8 +605,8 @@ openhost(
 		    ai->ai_addrlen) == -1)
 #endif /* SYS_VXWORKS */
 	    error("connect", "", "");
-	if (ai != NULL)
-		freeaddrinfo(ai);
+
+	freeaddrinfo(ai);
 	havehost = 1;
 	req_pkt_size = REQ_LEN_NOMAC;
 	impl_ver = IMPL_XNTPD;
@@ -634,11 +620,11 @@ openhost(
  */
 static int
 sendpkt(
-	char *xdata,
-	int xdatalen
+	void *	xdata,
+	size_t	xdatalen
 	)
 {
-	if (send(sockfd, xdata, (size_t)xdatalen, 0) == -1) {
+	if (send(sockfd, xdata, xdatalen, 0) == -1) {
 		warning("write to %s failed", currenthost, "");
 		return -1;
 	}
@@ -654,11 +640,7 @@ static void
 growpktdata(void)
 {
 	pktdatasize += INCDATASIZE;
-	pktdata = (char *)realloc(pktdata, (unsigned)pktdatasize);
-	if (pktdata == 0) {
-		(void) fprintf(stderr, "%s: realloc() failed!\n", progname);
-		exit(1);
-	}
+	pktdata = erealloc(pktdata, (size_t)pktdatasize);
 }
 
 
@@ -676,7 +658,7 @@ getresponse(
 	)
 {
 	struct resp_pkt rpkt;
-	struct timeval tvo;
+	struct sock_timeval tvo;
 	int items;
 	int i;
 	int size;
@@ -710,9 +692,9 @@ getresponse(
 
     again:
 	if (firstpkt)
-	    tvo = tvout;
+		tvo = tvout;
 	else
-	    tvo = tvsout;
+		tvo = tvsout;
 	
 	FD_SET(sockfd, &fds);
 	n = select(sockfd+1, &fds, (fd_set *)0, (fd_set *)0, &tvo);
@@ -823,7 +805,8 @@ getresponse(
 		pad = esize - size;
 	else 
 		pad = 0;
-	if ((datasize = items*size) > (n-RESP_HEADER_SIZE)) {
+	datasize = items * size;
+	if ((size_t)datasize > (n-RESP_HEADER_SIZE)) {
 		if (debug)
 		    printf(
 			    "Received items %d, size %d (total %d), data in packet is %d\n",
@@ -869,7 +852,7 @@ getresponse(
 	if ((datap + datasize + (pad * items)) > (pktdata + pktdatasize)) {
 		int offset = datap - pktdata;
 		growpktdata();
-	        *rdata = pktdata; /* might have been realloced ! */
+		*rdata = pktdata; /* might have been realloced ! */
 		datap = pktdata + offset;
 	}
 	/* 
@@ -877,9 +860,9 @@ getresponse(
 	 * items.  This is so we can play nice with older implementations
 	 */
 
-	tmp_data = (char *)rpkt.data;
-	for(i = 0; i <items; i++){
-		memmove(datap, tmp_data, (unsigned)size);
+	tmp_data = rpkt.data;
+	for (i = 0; i < items; i++) {
+		memcpy(datap, tmp_data, (unsigned)size);
 		tmp_data += size;
 		memset(datap + size, 0, pad);
 		datap += size + pad;
@@ -897,36 +880,63 @@ getresponse(
 	 */
 	++numrecv;
 	if (numrecv <= lastseq)
-	    goto again;
+		goto again;
 	return INFO_OKAY;
 }
 
 
 /*
  * sendrequest - format and send a request packet
+ *
+ * Historically, ntpdc has used a fixed-size request packet regardless
+ * of the actual payload size.  When authenticating, the timestamp, key
+ * ID, and digest have been placed just before the end of the packet.
+ * With the introduction in late 2009 of support for authenticated
+ * ntpdc requests using larger 20-octet digests (vs. 16 for MD5), we
+ * come up four bytes short.
+ *
+ * To maintain interop while allowing for larger digests, the behavior
+ * is unchanged when using 16-octet digests.  For larger digests, the
+ * timestamp, key ID, and digest are placed immediately following the
+ * request payload, with the overall packet size variable.  ntpd can
+ * distinguish 16-octet digests by the overall request size being
+ * REQ_LEN_NOMAC + 4 + 16 with the auth bit enabled.  When using a
+ * longer digest, that request size should be avoided.
+ *
+ * With the form used with 20-octet and larger digests, the timestamp,
+ * key ID, and digest are located by ntpd relative to the start of the
+ * packet, and the size of the digest is then implied by the packet
+ * size.
  */
 static int
 sendrequest(
 	int implcode,
 	int reqcode,
 	int auth,
-	int qitems,
-	int qsize,
+	u_int qitems,
+	size_t qsize,
 	char *qdata
 	)
 {
 	struct req_pkt qpkt;
-	int datasize;
+	size_t	datasize;
+	size_t	reqsize;
+	u_long	key_id;
+	l_fp	ts;
+	l_fp *	ptstamp;
+	int	maclen;
+	char	pass_prompt[32];
+	char *	pass;
 
-	memset((char *)&qpkt, 0, sizeof qpkt);
+	memset(&qpkt, 0, sizeof(qpkt));
 
 	qpkt.rm_vn_mode = RM_VN_MODE(0, 0, 0);
 	qpkt.implementation = (u_char)implcode;
 	qpkt.request = (u_char)reqcode;
 
 	datasize = qitems * qsize;
-	if (datasize != 0 && qdata != NULL) {
-		memmove((char *)qpkt.data, qdata, (unsigned)datasize);
+	if (datasize && qdata != NULL) {
+		memcpy(qpkt.data, qdata, datasize);
 		qpkt.err_nitems = ERR_NITEMS(0, qitems);
 		qpkt.mbz_itemsize = MBZ_ITEMSIZE(qsize);
 	} else {
@@ -936,53 +946,64 @@ sendrequest(
 
 	if (!auth || (keyid_entered && info_auth_keyid == 0)) {
 		qpkt.auth_seq = AUTH_SEQ(0, 0);
-		return sendpkt((char *)&qpkt, req_pkt_size);
-	} else {
-		l_fp ts;
-		int maclen = 0;
-		const char *pass = "\0";
-		struct req_pkt_tail *qpktail;
-
-		qpktail = (struct req_pkt_tail *)((char *)&qpkt + req_pkt_size
-		    + MAX_MAC_LEN - sizeof(struct req_pkt_tail));
-
-		if (info_auth_keyid == 0) {
-			if (((struct conf_peer *)qpkt.data)->keyid > 0)
-				info_auth_keyid = ((struct conf_peer *)qpkt.data)->keyid;
-			else {
-				maclen = getkeyid("Keyid: ");
-				if (maclen == 0) {
-					(void) fprintf(stderr,
-					    "Invalid key identifier\n");
-					return 1;
-				}
-				info_auth_keyid = maclen;
-			}
-		}
-		if (!authistrusted(info_auth_keyid)) {
-			pass = getpass("MD5 Password: ");
-			if (*pass == '\0') {
-				(void) fprintf(stderr,
-				    "Invalid password\n");
-				return (1);
-			}
-		}
-		authusekey(info_auth_keyid, info_auth_keytype, (const u_char *)pass);
-		authtrust(info_auth_keyid, 1);
-		qpkt.auth_seq = AUTH_SEQ(1, 0);
-		qpktail->keyid = htonl(info_auth_keyid);
-		get_systime(&ts);
-		L_ADD(&ts, &delay_time);
-		HTONL_FP(&ts, &qpktail->tstamp);
-		maclen = authencrypt(info_auth_keyid, (u_int32 *)&qpkt,
-		    req_pkt_size);
-		if (maclen == 0) {  
-			(void) fprintf(stderr, "Key not found\n");
-			return (1);
-		}
-		return sendpkt((char *)&qpkt, (int)(req_pkt_size + maclen));
+		return sendpkt(&qpkt, req_pkt_size);
 	}
-	/*NOTREACHED*/
+
+	if (info_auth_keyid == 0) {
+		key_id = getkeyid("Keyid: ");
+		if (!key_id) {
+			fprintf(stderr, "Invalid key identifier\n");
+			return 1;
+		}
+		info_auth_keyid = key_id;
+	}
+	if (!authistrusted(info_auth_keyid)) {
+		snprintf(pass_prompt, sizeof(pass_prompt),
+			 "%s Password: ",
+			 keytype_name(info_auth_keytype));
+		pass = getpass(pass_prompt);
+		if ('\0' == pass[0]) {
+			fprintf(stderr, "Invalid password\n");
+			return 1;
+		}
+		authusekey(info_auth_keyid, info_auth_keytype,
+			   (u_char *)pass);
+		authtrust(info_auth_keyid, 1);
+	}
+	qpkt.auth_seq = AUTH_SEQ(1, 0);
+	if (info_auth_hashlen > 16) {
+		/*
+		 * Only ntpd which expects REQ_LEN_NOMAC plus maclen
+		 * octets in an authenticated request using a 16 octet
+		 * digest (that is, a newer ntpd) will handle digests
+		 * larger than 16 octets, so for longer digests, do
+		 * not attempt to shorten the requests for downlevel
+		 * ntpd compatibility.
+		 */
+		if (REQ_LEN_NOMAC != req_pkt_size)
+			return 1;
+		reqsize = REQ_LEN_HDR + datasize + sizeof(*ptstamp);
+		/* align to 32 bits */
+		reqsize = (reqsize + 3) & ~3;
+	} else
+		reqsize = req_pkt_size;
+	ptstamp = (void *)((char *)&qpkt + reqsize);
+	ptstamp--;
+	get_systime(&ts);
+	L_ADD(&ts, &delay_time);
+	HTONL_FP(&ts, ptstamp);
+	maclen = authencrypt(info_auth_keyid, (void *)&qpkt, reqsize);
+	if (!maclen) {  
+		fprintf(stderr, "Key not found\n");
+		return 1;
+	} else if (maclen != (info_auth_hashlen + sizeof(keyid_t))) {
+		fprintf(stderr,
+			"%d octet MAC, %u expected with %u octet digest\n",
+			maclen, (info_auth_hashlen + sizeof(keyid_t)),
+			info_auth_hashlen);
+		return 1;
+	}
+	return sendpkt(&qpkt, reqsize + maclen);
 }
 
 
@@ -1007,7 +1028,7 @@ doquery(
 	int res;
 	char junk[512];
 	fd_set fds;
-	struct timeval tvzero;
+	struct sock_timeval tvzero;
 
 	/*
 	 * Check to make sure host is open
@@ -1041,7 +1062,7 @@ again:
 	 */
 	res = sendrequest(implcode, reqcode, auth, qitems, qsize, qdata);
 	if (res != 0)
-	    return res;
+		return res;
 	
 	/*
 	 * Get the response.  If we got a standard error, print a message
@@ -1079,36 +1100,36 @@ again:
  	/* log error message if not told to be quiet */
  	if ((res > 0) && (((1 << res) & quiet_mask) == 0)) {
 		switch(res) {
-		    case INFO_ERR_IMPL:
+		case INFO_ERR_IMPL:
 			/* Give us a chance to try the older implementation. */
 			if (implcode == IMPL_XNTPD)
 				break;
 			(void) fprintf(stderr,
 				       "***Server implementation incompatable with our own\n");
 			break;
-		    case INFO_ERR_REQ:
+		case INFO_ERR_REQ:
 			(void) fprintf(stderr,
 				       "***Server doesn't implement this request\n");
 			break;
-		    case INFO_ERR_FMT:
+		case INFO_ERR_FMT:
 			(void) fprintf(stderr,
 				       "***Server reports a format error in the received packet (shouldn't happen)\n");
 			break;
-		    case INFO_ERR_NODATA:
+		case INFO_ERR_NODATA:
 			(void) fprintf(stderr,
 				       "***Server reports data not found\n");
 			break;
-		    case INFO_ERR_AUTH:
+		case INFO_ERR_AUTH:
 			(void) fprintf(stderr, "***Permission denied\n");
 			break;
-		    case ERR_TIMEOUT:
+		case ERR_TIMEOUT:
 			(void) fprintf(stderr, "***Request timed out\n");
 			break;
-		    case ERR_INCOMPLETE:
+		case ERR_INCOMPLETE:
 			(void) fprintf(stderr,
 				       "***Response from server was incomplete\n");
 			break;
-		    default:
+		default:
 			(void) fprintf(stderr,
 				       "***Server returns unknown error code %d\n", res);
 			break;
@@ -1124,33 +1145,20 @@ again:
 static void
 getcmds(void)
 {
-#if defined(HAVE_LIBREADLINE) || defined(HAVE_LIBEDIT)
-	char *line;
+	char *	line;
+	int	count;
+
+	ntp_readline_init(interactive ? prompt : NULL);
 
 	for (;;) {
-		if ((line = readline(interactive?prompt:"")) == NULL) return;
-		if (*line) add_history(line);
+		line = ntp_readline(&count);
+		if (NULL == line)
+			break;
 		docmd(line);
 		free(line);
 	}
-#else /* not (HAVE_LIBREADLINE || HAVE_LIBEDIT) */
-	char line[MAXLINE];
 
-	for (;;) {
-		if (interactive) {
-#ifdef VMS	/* work around a problem with mixing stdout & stderr */
-			fputs("",stdout);
-#endif
-			(void) fputs(prompt, stderr);
-			(void) fflush(stderr);
-		}
-
-		if (fgets(line, sizeof line, stdin) == NULL)
-		    return;
-
-		docmd(line);
-	}
-#endif /* not HAVE_LIBREADLINE || HAVE_LIBEDIT */
+	ntp_readline_uninit();
 }
 
 
@@ -1482,7 +1490,7 @@ getarg(
 static int
 getnetnum(
 	const char *hname,
-	struct sockaddr_storage *num,
+	sockaddr_u *num,
 	char *fullhost,
 	int af
 	)
@@ -1490,9 +1498,7 @@ getnetnum(
 	int sockaddr_len;
 	struct addrinfo hints, *ai = NULL;
 
-	sockaddr_len = (af == AF_INET)
-			   ? sizeof(struct sockaddr_in)
-			   : sizeof(struct sockaddr_in6);
+	sockaddr_len = SIZEOF_SOCKADDR(af);
 	memset((char *)&hints, 0, sizeof(struct addrinfo));
 	hints.ai_flags = AI_CANONNAME;
 #ifdef AI_ADDRCONFIG
@@ -1502,7 +1508,7 @@ getnetnum(
 	/* decodenetnum only works with addresses */
 	if (decodenetnum(hname, num)) {
 		if (fullhost != 0) {
-			getnameinfo((struct sockaddr *)num, sockaddr_len, 
+			getnameinfo(&num->sa, sockaddr_len, 
 				    fullhost, sizeof(fullhost), NULL, 0, 
 				    NI_NUMERICHOST); 
 		}
@@ -1525,13 +1531,13 @@ getnetnum(
  */
 char *
 nntohost(
-	struct sockaddr_storage *netnum
+	sockaddr_u *netnum
 	)
 {
 	if (!showhostnames)
-	    return stoa(netnum);
+		return stoa(netnum);
 
-	if ((netnum->ss_family == AF_INET) && ISREFCLOCKADR(netnum))
+	if (ISREFCLOCKADR(netnum))
 		return refnumtoa(netnum);
 	return socktohost(netnum);
 }
@@ -1799,23 +1805,36 @@ keytype(
 	FILE *fp
 	)
 {
-	if (pcmd->nargs == 0)
-	    fprintf(fp, "keytype is %s\n",
-		    (info_auth_keytype == KEY_TYPE_MD5) ? "MD5" : "???");
-	else
-	    switch (*(pcmd->argval[0].string)) {
-		case 'm':
-		case 'M':
-		    info_auth_keytype = KEY_TYPE_MD5;
-		    break;
+	const char *	digest_name;
+	size_t		digest_len;
+	int		key_type;
 
-		default:
-		    fprintf(fp, "keytype must be 'md5'\n");
-	    }
+	if (!pcmd->nargs) {
+		fprintf(fp, "keytype is %s with %u octet digests\n",
+			keytype_name(info_auth_keytype),
+			info_auth_hashlen);
+		return;
+	}
+
+	digest_name = pcmd->argval[0].string;
+	digest_len = 0;
+	key_type = keytype_from_text(digest_name, &digest_len);
+
+	if (!key_type) {
+		fprintf(fp, "keytype must be 'md5'%s\n",
+#ifdef OPENSSL
+			" or a digest type provided by OpenSSL");
+#else
+			"");
+#endif
+		return;
+	}
+
+	info_auth_keytype = key_type;
+	info_auth_hashlen = digest_len;
 }
 
 
-
 /*
  * passwd - get an authentication key
  */
@@ -1982,7 +2001,7 @@ getkeyid(
 #ifndef SYS_WINNT
 	if ((fi = fdopen(open("/dev/tty", 2), "r")) == NULL)
 #else
-	    if ((fi = _fdopen((int)GetStdHandle(STD_INPUT_HANDLE), "r")) == NULL)
+	if ((fi = _fdopen(open("CONIN$", _O_TEXT), "r")) == NULL)
 #endif /* SYS_WINNT */
 		fi = stdin;
 	    else
